# Note that a few functions have already been defined
# in 03pre-initialization.bash.

function pause_before_retrying {
    inform "to mitigate or avoid abuse of the service over the network,"\
     "pausing $SLEEP_DURATION seconds before retrying..."
    inform "(press ctrl-C to abandon the attempt and exit the program)"
    sleep $SLEEP_DURATION
    inform "retrying now..."
}

# usage: sort_by_filename [--no-sums] FILE...
#
# The FILEs list the filenames, usually preceded by checksums.
function sort_by_filename {
    local SBF_DO_SWAP=1; [ "$1" = '--no-sums' ] && shift && SBF_DO_SWAP=0
    local SBF_SED_SCRIPT='s/^(\S+)(\s+)(\S+)\s*$/\3\2\1/'
    if (($SBF_DO_SWAP)); then
        sed -re "$SBF_SED_SCRIPT" $@ | sort | uniq\
        | sed -re "$SBF_SED_SCRIPT"
    else
        sort $@ | uniq
    fi
}

# Some commands can accept only a limited number of arguments on their command
# lines, yet it can be inefficient to invoke such a command per argument. As a
# compromise, this function invokes a command once per $BATCH_SIZE arguments.
# (If the number of arguments is moderate, not exceeding $BATCH_SIZE, then this
# function merely invokes the command once in the normal way.)
#
# The ARGS-FILE is to give one argument per line.
#
# (GNU split *might* instead have been used. Whether that would have been
# better is unknown.)
#
# usage: run_by_batch [--persist] CMD PRE-ARGS ARGS-FILE POST-ARGS [PREFIX]
function run_by_batch {
    local RBB_DO_PERSIST=0; [ $1 = '--persist' ] && shift && RBB_DO_PERSIST=1
    local RBB_DATA="$(mktemp)"
    local RBB_BATCH_DATA="$(mktemp)"
    local RBB_EXTRA_DATA="$(mktemp)"
    local RBB_SIZE
    local RBB_SIZE1
    cat $3 >"$RBB_EXTRA_DATA"
    while true; do
        cat "$RBB_EXTRA_DATA" >"$RBB_DATA"
        RBB_SIZE=$(cat "$RBB_DATA" | wc -l)
        RBB_SIZE1=$RBB_SIZE;\
         (($RBB_SIZE1 > $BATCH_SIZE)) && RBB_SIZE1=$BATCH_SIZE
        (($RBB_SIZE1)) || break
        head -n${RBB_SIZE1} "$RBB_DATA" >"$RBB_BATCH_DATA"
        tail -n+$(($RBB_SIZE1 + 1)) "$RBB_DATA" >"$RBB_EXTRA_DATA"
        inform "running $1 on a $RBB_SIZE1-record batch"\
         "of the $RBB_SIZE remaining records, the batch consisting of these:"
        print_bar
        for A in $(<"$RBB_BATCH_DATA"); do echo "$A"; done
        print_bar
        until {
            "$1" $2 $(
                echo -n "$5"
                cat "$RBB_BATCH_DATA"
            ) $4
        } || ((!$RBB_DO_PERSIST)); do pause_before_retrying; done
        print_bar
    done
    rm -f "$RBB_DATA" "$RBB_BATCH_DATA" "$RBB_EXTRA_DATA"
}

# This function runs wget insistently. Should wget fail,
# files are deleted and wget is tried again.
#
# usage: wget_insist OPTIONS FILES
#
# The OPTIONS must be passed as a single string.
function wget_insist {
    local WGET2_OPTIONS="$1" && shift
    until wget $WGET2_OPTIONS "$@"; do
        rm -f$V $(find "$@" -maxdepth 0 -type f)
        pause_before_retrying
    done
}

function expand_languages {
    local EL_PAT='\$\{LANGS\[\*\]\}'
    local EL_LANG
    local EL_RECORD
    for EL_RECORD in "$@"; do
        if { echo "$EL_RECORD" | grep -qE "${EL_PAT}"; }; then
            for EL_LANG in "${LANGS[@]}"; do
                echo "$EL_RECORD" | sed -re "s/${EL_PAT}/${EL_LANG}/"
            done
        else
            echo "$EL_RECORD"
        fi
    done
}

function decompress {
    local DECOMPRESS_DIR
    local DECOMPRESS_BASE
    local DECOMPRESS_STEM
    local DECOMPRESS_EXT
    local DECOMPRESS_CMD
    local DECOMPRESS_FILEPATH
    for DECOMPRESS_FILEPATH in "$@"; do
        DECOMPRESS_DIR="$(dirname "$DECOMPRESS_FILEPATH")"
        DECOMPRESS_BASE="$(basename "$DECOMPRESS_FILEPATH")"
        DECOMPRESS_STEM="${DECOMPRESS_BASE%.+([^/.])}"
        DECOMPRESS_EXT="${DECOMPRESS_BASE#${DECOMPRESS_STEM}.}"
        DECOMPRESS_CMD='false'
        case "$DECOMPRESS_EXT" in
            xz)  DECOMPRESS_CMD='xzcat';;
            bz2) DECOMPRESS_CMD='bzcat';;
            gz)  DECOMPRESS_CMD='zcat';;
            *)
                 die "the filename $DECOMPRESS_FILEPATH"\
                 "indicates no decompression method of which"\
                 "this program is capable"
        esac
        pushd >/dev/null "$TARGET/$DECOMPRESS_DIR"
        "$DECOMPRESS_CMD" >"$DECOMPRESS_STEM" "$DECOMPRESS_BASE"
        touch -r "$DECOMPRESS_BASE" "$DECOMPRESS_STEM"
        popd >/dev/null
    done
}

# (Here is a utility function for is_ctrl_wanted.)
# example: icw_stem "i18n/Translation-@"  -> "i18n/Translation-+([^/.-])"
# example: icw_stem "i18n/Translation-en" -> "i18n/Translation-en"
function icw_stem {
    if [[ "$1" = *-@ ]]; then
        echo "${1%@}"'+([^/.-])'
    else
        echo "$1"
    fi
}

# usage: is_ctrl_wanted WANTED-LISTFILE CTRL
# example: is_ctrl_wanted per-area-ctrl.list main/i18n/Translation-fr.bz2
function is_ctrl_wanted {
    local ICW_STEM
    local ICW_PATTERN
    for ICW_PATTERN in $(expand_languages $(<"$LIST_DIR/$1")); do
        if [[ "$ICW_PATTERN" = *.@ ]]; then
            ICW_STEM="$(icw_stem "$(eval echo "${ICW_PATTERN%.@}")")"
            if (($OPT_SUMS)); then
                [[ "$2" = *([^/.])/${ICW_STEM}?(.@($COMPRESS_EXT)) ]]\
                 && return 0
            else
                [[ "$2" = *([^/.])/${ICW_STEM}.@($COMPRESS_EXT) ]]\
                 && return 0
            fi
        else
            ICW_STEM="$(icw_stem "$(eval echo "$ICW_PATTERN")")"
            [[ "$2" = *([^/.])/${ICW_STEM} ]] && return 0
        fi
    done
    return 1;
}

# usage: filepaths_from_sources_ctrl CTRL AREA
function filepaths_from_sources_ctrl {
    local FSC_P1='^Checksums-Sha256:'
    local FSC_P2a='(\s+\S+){2}\s+'
    if (($OPT_SUMS)); then
        FSC_P2a='\s+(\S+)\s+\S+\s+'
    fi
    local FSC_P2b='((lib)?[[:alnum:]])([[:alnum:].+-]*)(_\S+)\s*'
    local FSC_P2='^'"${FSC_P2a}${FSC_P2b}"'$'
    local FSC_R2='pool/'"$2"'/\2/\2\4/\2\4\5'
    local FSC_G1=':a;/'
    local FSC_G2='/bb;d;:b;n;s%'
    local FSC_G3s='%\1  '
    local FSC_G3='%'
    local FSC_G4='%p;tb;ba'
    local FSC_LEFT="${FSC_G1}${FSC_P1}${FSC_G2}${FSC_P2}"
    local FSC_RIGHT="${FSC_R2}${FSC_G4}"
    if (($OPT_SUMS)); then
        sed -rne "${FSC_LEFT}${FSC_G3s}${FSC_RIGHT}" <"$1"
    else
        sed -rne "${FSC_LEFT}${FSC_G3}${FSC_RIGHT}" <"$1"
    fi
}

# usage: filepaths_from_packages_ctrl CTRL
function filepaths_from_packages_ctrl {
    if (($OPT_SUMS)); then
        sed -rne 's/^(Filename|SHA256):\s*(\S+).*$/\2/p'\
        <"$1" | sed -rne 'h;n;G;s/\n/  /;p'
    else
        sed -rne 's/^Filename:\s*(\S+).*$/\1/p' <"$1"
    fi
}

# usage: filepaths_from_sha_ctrl CTRL DIR
function filepaths_from_sha_ctrl {
    if (($OPT_SUMS)); then
        sed -re 's%^(\S+)\s+\./%\1  '"$2"'/%' <"$1"
    else
        sed -re 's%^\S+\s+\./%'"$2"'/%' <"$1"
    fi
}

# usage: does_need_extra_zzz FILEPATH CTRL
#
# FILEPATH omits the first two dirs, for example main/i18n rather than
# dists/buster/main/i18n.
function does_need_extra_zzz {
    local DNEZ_PATTERN
    for DNEZ_PATTERN in $(<"$2"); do
        [[ "$1" = $(eval echo "$DNEZ_PATTERN") ]] && return 0;
    done
    return 1;
}

# Using the syntax of ln(1), symbolically link, and then touch the symlink
# to match symlink's timestamp to that of its target. Example:
#
#     ln_touch existing foo/mylink
#
# The result is a new symlink foo/mylink that points to the existing
# file foo/existing and shares the latter's timestamp.
function ln_touch {
    local LNT_TARGET
    if [[ "$1" = /* ]]; then
        LNT_TARGET="$1"
    else
        LNT_TARGET="$(dirname "$2")/$1"
    fi
    ln -s$V "$1" "$2"
    touch -hr "$LNT_TARGET" "$2"
}

# The maintainer could think of no suitably short name for this function other
# than cat2, but the function acts as cat with an optional filter. The filter
# engages when OPT_SMALL_TEST is selected, blocking most package pathnames but
# passing a few for test retrieval.
function cat2 {
    if (($OPT_SMALL_TEST)); then
        grep -h "/${SMALL_TEST}/" "$@"
    else
        cat "$@"
    fi
}

# usage: stamp_per_release STAMP-FILE RELEASE-FILE
function stamp_per_release {
    touch -d "$(sed -rne 's/^Date:\s*//p;T;q' "$2")" "$1"
}

# usage: epoch_per_stamp STAMP-FILE
function epoch_per_stamp { stat -c%Y "$1"; }

# usage: ddate_per_stamp STAMP-FILE
function ddate_per_stamp { stat -c%y "$1"; }

# usage: month_per_stamp STAMP-FILE
function month_per_stamp {
    stat -c%y "$1" | sed -re 's/^([[:digit:]]{4})-([[:digit:]]{2})-.*$/\1\2/'
}

# usage: convert_format_ss_to_d2 DATE-IN-SNAPSHOT-FORMAT
#
# Return the date in a format the date(1) command can read. (It's called _d2
# rather than _d because it is not the date command's canonical format.)
CFSSD_PAT1='^([[:digit:]]{8})T([[:digit:]]{2})'
CFSSD_PAT2='([[:digit:]]{2})([[:digit:]]{2})Z$'
CFSSD_PAT="${CFSSD_PAT1}${CFSSD_PAT2}"
function convert_format_ss_to_d2 {
     echo "$1" | sed -re "s/${CFSSD_PAT}/\\1 \\2:\\3:\\4 UTC/"
}

# Take one or more lines of whitespace-separated words on stdin; return a neat
# column of words on stdout. Extra space and blank lines on stdin have no
# effect: the function strips them.
function columnate { sed -re 's/^\s*(.*\S)\s*$/\1/;ta;d;:a;s/\s+/\n/g;'; }

# Note: as presently modified, this function is believed to be safe to use even
# outside the pool because it avoids directory links. Inside the pool, the
# function assumes that no disruptive directory links are involved and, thus,
# skips the safety check.
function copy_locally_available {
    local CLA_DO
    local CLA_FILEPATH
    for CLA_FILEPATH in $(<"$1"); do
        if
            {
                ((!$OPT_SMALL_TEST))\
                || [[ "$CLA_FILEPATH" = */${SMALL_TEST}/* ]]
            }\
            && ! [ -e "$TARGET/$CLA_FILEPATH" ]\
            && [ -f "$OLD/$CLA_FILEPATH" ]\
            && {
                [[ "$CLA_FILEPATH" = pool/* ]]\
                || [ "$(realpath -e "$OLD/$CLA_FILEPATH")"\
                 = "$OLD/$CLA_FILEPATH" ]
            }
        then
            mkdir -p$V $(dirname "$TARGET/$CLA_FILEPATH")
            if (($OPT_HARDLINK)); then
                ln $VV  "$OLD/$CLA_FILEPATH" "$TARGET/$CLA_FILEPATH"
            else
                cp -a$V "$OLD/$CLA_FILEPATH" "$TARGET/$CLA_FILEPATH"
            fi
        fi
    done
}

function inform_of_scan {
    inform "scanning $1..."
}

