# (Here is a utility function for the use of the recursive fetcher. For the
# fetcher itself, see below.)
#
# usage: inform_of_intent_to_recursively_fetch PROTOCOL
RF_FILES_TO_RSYNC=''
function inform_of_intent_to_recursively_fetch {
    inform "preparing to fetch the following via $1:"
    print_bar
    for A in $RF_FILES_TO_RSYNC; do echo "$A"; done
    print_bar
}

# This function, the recursive fetcher, is meant to be called from phase C
# or D, where the variables MIRROR, IS_BACKPORTS and DIST1 will already have
# been set. It also uses SNAPSHOT_SSDATE from phase B if IS_BACKPORTS is true.
# (Such data connections are more haphazard and more brittle than one would
# like. They rely too much on comments like the one you are reading. They carry
# over from an earlier era in which the program was smaller. Now that the
# program has grown, a reorganization of the data connections, along perhaps
# with a redesign of the phase selector of 05option-processing.bash, might be
# wanted at some future date.) The function is for fetching nonpool files only.
# It will likely need updating once Debian's next major version arrives.
#
# The "RFP" stands for "Recursive Filepath."
# The "RFPS", plural, stands for "Recursive Filepaths."
# The "RRFP" stands for "Relative Recursive Filepath."
# The "ZFP" stands for "Decompressible Filepath."
# The "ZFPS", plural, stands for "Decompressible Filepaths."
# The "FP" in general stands for "Filepath,"
# so for instance "ROOTFPS" stands for "Root Filepaths."
#
# The function is recursive insofar as it
#
#   * fetches files and symlinks (making directories for them as necessary),
#   * lists the files and symlinks at which the just-fetched symlinks point,
#     and then
#   * calls itself to fetch the files and symlinks the last list lists.
#
# The recursion's purpose is to follow symlinks. The recursion bottoms out
# after a cycle during which no symlinks but only files are fetched.
#
# example: recursively_fetch\
#  dists/$DIST_BACKPORTS/Release.gpg\
#  dists/$DIST/main/i18n/Translation-fr.bz2
#
# Note that the arguments to this function begin two levels higher in the
# control directory hierarchy than does the last argument to is_ctrl_wanted.
#
# usage: recursively fetch FILEPATH-TO-FETCH [FILEPATH-TO-FETCH...]
function recursively_fetch {
    (($OPT_SUMS)) && die "impossible: recursively_fetch called while -s is set"
    local RF_FILEPATH
    local RF_IS_FIRST
    local RF_RRFP
    local RF_RFPS=''
    RF_FILES_TO_RSYNC="$(
        RF_IS_FIRST=1
        for RF_FILEPATH in "$@"; do
            if (($IS_BACKPORTS)); then echo -n\
             "https://$MIRROR/$ARCHIVE_NAME/$SNAPSHOT_SSDATE/$RF_FILEPATH "
            else
                if (($RF_IS_FIRST)); then
                    echo -n "$MIRROR" && RF_IS_FIRST=0
                fi
                echo -n "::$ARCHIVE_NAME/$RF_FILEPATH "
            fi
        done
    )"
    if [ -n "$RF_FILES_TO_RSYNC" ]; then
        if (($IS_BACKPORTS)); then
            inform_of_intent_to_recursively_fetch 'wget'
            # Transferring via https, wget apparently, unfortunately is unable
            # to preserve symlinks. Therefore, though the data is transferred
            # and safely stored, the recursion fails.
            #  -nc       do not clobber existing files
            #  -x        download not into current dir but into hierarchy
            #  -nH       do not generate hostname dirs
            #  -N        update existing files if their timestamps have changed
            # --cut-dirs cut leading dirs
            #  -P        root the target hierarchy
            # The --cut-dirs=5 strips five levels from the likes of
            # archive/debian/19700101T000000Z/dists/$DISTS_BACKPORTS/main/...
            # leaving just main/...
            wget_insist\
             "-nc -x -nH --cut-dirs=5 -P $TARGET/dists/$DIST_BACKPORTS"\
             $RF_FILES_TO_RSYNC
            #
            # The remainder of the action below fixes backports, recovering
            # the failed recursion. Because wget, upon which the retrieval of
            # backports relies, apparently cannot retrieve symlinks as
            # symlinks via http (presumably because http does not support such
            # retrieval), backports is made below to imitate the symlink
            # structure of stable.
            #
            # Following pattern, one might have expected
            # zzz-dists/stable-backports next but the archive does not do that.
            ZZZDIR="$TARGET/zzz-dists/$DIST1"
            mkdir -pv "$ZZZDIR"
            for FP1 in $(
                find "$TARGET/dists/$DIST1" -maxdepth 1 -type f -printf '%P\n'
            ); do
                # Note the $DIST, rather than $DIST1, on the next line.
                if [ -h "$TARGET/dists/$DIST/$FP1" ]; then
                    mv -v "$TARGET/dists/$DIST1/$FP1" "$ZZZDIR/$FP1"
                    ln_touch "../../zzz-dists/$DIST1/$FP1"\
                     "$TARGET/dists/$DIST1/$FP1"
                fi
            done
            for FP1 in $(
                find "$TARGET/dists/$DIST1" -mindepth 2 -type f -printf '%P\n'
            ); do
                # Note the $DIST, rather than $DIST1, on the next line.
                if {
                    [ -h "$TARGET/dists/$DIST/$FP1" ] ||\
                    does_need_extra_zzz "$FP1" "$LIST_DIR/extra-zzz.list";
                }; then
                    CHECKSUM="$(
                        sha256sum "$TARGET/dists/$DIST1/$FP1"\
                        | sed -re 's/\s.*//'
                    )"
                    SHORTDIR='by-hash/SHA256'
                    HASHDIR="$(dirname "$TARGET/dists/$DIST1/$FP1")/$SHORTDIR"
                    mkdir -pv "$HASHDIR"
                    mv -v "$TARGET/dists/$DIST1/$FP1" "$HASHDIR/$CHECKSUM"
                    ln_touch "$SHORTDIR/$CHECKSUM" "$TARGET/dists/$DIST1/$FP1"
                fi
            done
        else
            inform_of_intent_to_recursively_fetch 'rsync'
            # Unlike wget, rsync can preserve symlinks.
            # (The below option descriptions are verbatim from man 1 rsync.)
            #  -l        copy symlinks as symlinks
            #  -t        preserve modification times
            #  -p        preserve permissions
            #  -R        use relative path names
            #  -v        be verbose
            # --progress show progress during transfer
            # --stats    give some file-transfer stats
            until
                rsync -lKtRv --progress --stats\
                $RF_FILES_TO_RSYNC "$TARGET/"
            do pause_before_retrying; done
            for RF_FILEPATH in "$@"; do
                if [ -h "$TARGET/$RF_FILEPATH" ]; then
                    RF_RRFP="$(readlink "$TARGET/$RF_FILEPATH")"
                    if [[ "$RF_RRFP" = /* ]]; then
                        RF_RFPS="$RF_RFPS ${RF_RRFP#/}"
                    else
                        RF_RFPS="$RF_RFPS $(dirname "$RF_FILEPATH")/$RF_RRFP"
                    fi
                fi
            done
            [ -n "$RF_RFPS" ] && recursively_fetch $RF_RFPS
        fi
        print_bar
    fi
    true # would be necessary if the && were last
}

