# Determine the snapshot date to use.

# Though prepared in the form of a function's definition, this phase B is not
# called yet, but called during the loop that drives phases C and D.
#
# When called, the function sets these shell variables for external use:
#
#     SNAPSHOT_SSDATE usable in a URL at snapshot.debian.org
#     SNAPSHOT_D2DATE the same snapshot date, but in date(1)-compatible format
#
# The function also leaves these files in $TMPDIR/:
#
#     snapshot.stamp    the snapshot date again, but as a timestamp
#     phb-release.stamp for later comparison against release.stamp
#     phb-html-parses/* (possibly interesting)
#     phb-internal/*    (probably uninteresting)
#
# The variable DIST must already be set before the function is called.
# Also, the html file $TMPDIR/mirror-checks/$SNAPSHOT_TOC1 must already have
# been fetched.
#
# Various development notes follow.
#
# TRIPLICATION OF THE SNAPSHOT DATE
#
# Are three distinct means of returning a single snapshot date overkill?
# Answer: perhaps, but the triplicate results from the manner in which this
# program has evolved. The triplicate could be reduced/combined with modest
# programming effort if future development revealed that the triplicate posed a
# practical problem, but reduction/combination seems unnecessary at the moment.
#
# Conceptually, determining the right snapshot date should be pretty easy.
# A sufficiently experienced Debian user could do it manually in a few minutes
# by following a three-line instruction in English: "Find the latest backports
# release available on snapshot that precedes or equals the stable release. If
# two snapshots have captured the same release, prefer the later snapshot."
# However, because the task requires the parsing of human-readable web pages
# and for other reasons, teaching a shell script to complete the task
# automatically with a passably useful degree of reliability -- without
# invoking tools like Perl that would make it hard for future maintainers
# unfamiliar with the tools to maintain the script -- wants 200 lines or so of
# code, as follows. And even if Perl were used, the code would still not
# be short.
#
# DUPLICATION OF THE TIMESTAMP; CHOICE OF BASH AS A SCRIPTING LANGUAGE
#
# The phb-release.stamp is arguably unnecessary, since a release.stamp (without
# the phb-), which ought to be identical, should already exist. However, Bash
# as a scripting language is weakly scoped, lacks support for lazy
# instantiation, and has little support for compound objects of user-defined
# type, so the separate phb-release.stamp is kept to cause certain,
# inadvertently introduced potential future bugs to manifest themselves in a
# more obvious way. Without the separate phb-release.stamp, the program as
# presently designed feels too brittle.
#
# It could be argued that the last is sufficient reason to have used something
# other than Bash as a scripting language. Perhaps that argument is right, but
# the program seems not unlikely to require significant future maintenance on a
# two-year cycle (that is, upon each major release of Debian), possibly by
# persons other than the original author. Bash is already known to practically
# all potential future maintainers. Moreover, whatever Bash's
# shortcomings (whether real or perceived), Bash is especially suited to moving
# files around -- which, of course, is what this program chiefly does.
#
# Further notes on the Bash implementation are written at the end of the file.
#
# Meanwhile, the release.stamp duplication makes comparison of the duplicates
# necessary at the end of phase D, but that's all right. That's what catches
# the bugs. Indeed, it has already caught one, now fixed.
#
# IMPROPER BASH FUNCTIONS
#
# Unlike functions defined in the DEFINE FUNCTIONS section above, this function
# is meant to be called only once during the program's run. It is designed not
# as a proper function but merely as a device to code phase B, here in the
# source, separately from phases C and D. (Archaic programming languages
# like FORTRAN and BASIC would have called it a "subroutine" rather than
# a "function.")
#
SNAPSHOT_SSDATE=''
SNAPSHOT_D2DATE=''
function execute_phase_B {

    mkdir $VV $TMPDIR/phb-internal

    # Extract the stable release date from the corresponding Release file.
    stamp_per_release "$TMPDIR/phb-release.stamp"\
     "$TARGET/dists/$DIST/Release"
    local PHB_STABLE_RELEASE_MONTH="$(
        month_per_stamp "$TMPDIR/phb-release.stamp"
    )"

    # Phase A has already fetched the web
    # page, $TMPDIR/mirror-checks/$SNAPSHOT_TOC1, that lists the years and
    # months. Extract and list the years and months in plain text, as "197102"
    # for Feb. 1971.
    mkdir $VV "$TMPDIR/phb-html-parses"
    local PHB_PAT1='^[^<>]*<\s*a\s+href\s*=\s*"[^"]*year=([[:digit:]]+)[^"]*'
    local PHB_PAT2='month=([[:digit:]]+)[^"]*"\s*/?\s*>.*$'
    local PHB_PAT="${PHB_PAT1}${PHB_PAT2}"
    sed -rn\
     -e '\%<h1>[^<>]*</h1>%ba;d;:a;n;'\
     -e "s%${PHB_PAT}%\\1\\2%p;"\
     -e 'ba;'\
     "$TMPDIR/mirror-checks/$SNAPSHOT_TOC1"\
     | sed -re 's%^([[:digit:]]{4})([[:digit:]])$%\10\2%'\
     >"$TMPDIR/phb-html-parses/months.list"

    # Approximate overview of procedure:
    #
    #   * Read TOC2 for several months: $N_FLANKING_MONTHS before, one during,
    #     and $N_FLANKING_MONTHS after the stable release. (If a month after
    #     is unavailable, omit it.)
    #   * Compile a single list of snapshot dates from the months.
    #   * Scanning late to early, identify the latest snapshot that precedes or
    #     equals the stable release. (Die if that's the latest in the list.)
    #   * Step $SNAPSHOT_MARGIN snapshots older. (This step should be
    #     unnecessary except that snapshot's clock might be imperfectly
    #     synchronized with backports' clock.)
    #   * Starting from there, scanning early to late, find the earliest
    #     snapshot whose backports release succeeds the stable release. (Die
    #     if it's the earliest remaining.)
    #   * Step one snapshot older.
    #

    # In TOC1, index the month of the stable release (of, if that specific
    # month is not in TOC1, then the nearest earlier month). Record the index
    # in PHB_INDEX.
    local PHB_N_MONTHS=$(cat "$TMPDIR/phb-html-parses/months.list" | wc -l)
    (($PHB_N_MONTHS > 0)) || {
        die "the list of months given by $SNAPSHOT_SERVICE seems to be empty"
    }
    local PHB_INDEX=$PHB_N_MONTHS
    local PHB_MONTH1
    for PHB_MONTH1 in $(tac "$TMPDIR/phb-html-parses/months.list"); do
        [ $PHB_MONTH1 -le $PHB_STABLE_RELEASE_MONTH ] && break
        PHB_INDEX=$(($PHB_INDEX - 1))
        (($PHB_INDEX < 1)) && {
            die "$SNAPSHOT_SERVICE does not cover"\
            "the stable release month $PHB_STABLE_RELEASE_MONTH"
        }
    done
    (($PHB_INDEX > $N_FLANKING_MONTHS)) || {
        die "$SNAPSHOT_SERVICE covers too little before"\
        "the stable release month $PHB_STABLE_RELEASE_MONTH"
    }

    # Read the second-level tables of contents (TOC2) of nearby months.
    # Format and store the TOC2 of the nearby months together to
    # the single file "$TMPDIR/phb-html-parses/snapshots.list",
    # thereby compiling a list of candidate snapshot dates.
    local PHB_PERIOD_DURATION
    if (($PHB_INDEX + $N_FLANKING_MONTHS <= $PHB_N_MONTHS)); then
        PHB_PERIOD_DURATION=$((2*$N_FLANKING_MONTHS + 1))
    else
        PHB_PERIOD_DURATION=$((
            $PHB_N_MONTHS + $N_FLANKING_MONTHS + 1 - $PHB_INDEX
        ))
    fi
    touch "$TMPDIR/phb-html-parses/snapshots.list"
    PHB_PAT='^[^<>]*<\s*a\s+href\s*=\s*"([^"]+)"\s*/?\s*>.*$'
    local PHB_HMBS="https://$SNAPSHOT_SERVICE/$ARCHIVE_NAME"
    local PHB_YEARONLY
    local PHB_MONTHONLY
    # (Observe that PHB_MONTH1 has already been locally defined.)
    for PHB_MONTH1 in $(
        head -n$(($PHB_INDEX + $N_FLANKING_MONTHS))\
        "$TMPDIR/phb-html-parses/months.list"\
        | tail -n${PHB_PERIOD_DURATION}
    ); do
        PHB_YEARONLY="$(echo "$PHB_MONTH1" | sed -re 's/[[:digit:]]{2}$//')"
        PHB_MONTHONLY="$(echo "$PHB_MONTH1" | sed -re 's/^[[:digit:]]{4}//')"
        sed -rn\
         -e '\%<h1>[^<>]*</h1>%ba;d;:a;n;'\
         -e '\%<h2>[^<>]*</h2>%bb;ba;:b;n;'\
         -e "s%${PHB_PAT}%\\1%p;"\
         -e 'bb;'\
         <(wget -nH -O- "$PHB_HMBS/?year=$PHB_YEARONLY&month=$PHB_MONTHONLY")\
         | sed -rne 's%/$%%;/^[[:digit:]]{8}T/p;'\
         >>"$TMPDIR/phb-html-parses/snapshots.list"
    done
    # (Observe that the value of PHB_INDEX, indexing a month, is no longer
    # needed. Thus, the next segment can reuse the variable PHB_INDEX for a
    # different purpose.)

    # Index by $PHB_INDEX the latest snapshot that precedes or equals
    # the stable release.
    local PHB_N_SNAPSHOTS=$(
        cat "$TMPDIR/phb-html-parses/snapshots.list" | wc -l
    )
    (($PHB_N_SNAPSHOTS > 0)) || {
        die "$SNAPSHOT_SERVICE has no snapshots"\
        "during the months near the stable release"
    }
    PHB_INDEX=$PHB_N_SNAPSHOTS
    local PHB_SNAPSHOT1
    for PHB_SNAPSHOT1 in $(tac "$TMPDIR/phb-html-parses/snapshots.list"); do
        touch -d "$(convert_format_ss_to_d2 "$PHB_SNAPSHOT1")"\
         "$TMPDIR/phb-internal/stamp.stamp"
        [ "$TMPDIR/phb-internal/stamp.stamp" -nt "$TMPDIR/phb-release.stamp" ]\
         || break
        PHB_INDEX=$(($PHB_INDEX - 1))
        (($PHB_INDEX < 1)) && {
            die "$SNAPSHOT_SERVICE does not"\
            "seem to cover the stable release date and time"
        }
    done

    # Step $SNAPSHOT_MARGIN snapshots older.
    (($PHB_INDEX > $SNAPSHOT_MARGIN)) || {
        die "$SNAPSHOT_SERVICE seems"\
        "to have too few snapshots that precede the stable release"\
        "(it is possible, though unlikely, that"\
        "adding 1 to the value of N_FLANKING_MONTHS, or"\
        "subtracting 1 from the value of SNAPSHOT_MARGIN,"\
        "in the configuration file would fix the problem)"
    }
    PHB_INDEX=$(($PHB_INDEX - $SNAPSHOT_MARGIN))

    # Index by $PHB_JNDEX the earliest snapshot whose backports release
    # succeeds the stable release.
    local PHB_JNDEX=0
    local PHB_HAS_FOUND=0
    for PHB_SNAPSHOT1 in $(<"$TMPDIR/phb-html-parses/snapshots.list"); do
        PHB_JNDEX=$(($PHB_JNDEX + 1))
        if (($PHB_JNDEX >= $PHB_INDEX)); then
            wget -nH -O- >"$TMPDIR/phb-internal/wget-out.txt"\
             "$PHB_HMBS/$PHB_SNAPSHOT1/dists/$DIST_BACKPORTS/Release"
            stamp_per_release "$TMPDIR/phb-internal/stamp.stamp"\
             "$TMPDIR/phb-internal/wget-out.txt"
            [\
                    "$TMPDIR/phb-internal/stamp.stamp"\
                -nt "$TMPDIR/phb-release.stamp"\
            ] && {
                PHB_HAS_FOUND=1
                break
            }
        fi
    done
    (($PHB_HAS_FOUND)) || {
        die "$SNAPSHOT_SERVICE seems to have captured"\
        "no backports release that succeeds the stable release"\
        "(it is possible, though unlikely, that"\
        "adding 1 to the value of N_FLANKING_MONTHS"\
        "in the configuration file would fix the problem)"
    }
    (($PHB_JNDEX > $PHB_INDEX)) || {
        die "a least one backport is found that"\
        "succeeds the stable release, which is good, but none are found"\
        "that precede or equal it"\
        "(it is possible, though unlikely, that"\
        "adding 1 from the value of SNAPSHOT_MARGIN, or"\
        "adding 1 to the value of N_FLANKING_MONTHS,"\
        "in the configuration file would fix the problem)"
    }
    # Note that the last command guarantees that PHB_JNDEX can be reduced by 1
    # without risk of its shrinking to less than PHB_INDEX.

    # Step one snapshot older.
    PHB_JNDEX=$(($PHB_JNDEX - 1))

    # Set SNAPSHOT_SSDATE, SNAPSHOT_D2DATE, et al.
    local PHB_KNDEX=0
    PHB_HAS_FOUND=0 # reused
    for PHB_SNAPSHOT1 in $(<"$TMPDIR/phb-html-parses/snapshots.list"); do
        PHB_KNDEX=$(($PHB_KNDEX + 1))
        (($PHB_KNDEX >= $PHB_JNDEX)) && {
            PHB_HAS_FOUND=1
            break
        }
    done
    (($PHB_HAS_FOUND)) || {
        die "an internal glitch with indices has"\
        "arisen while setting SNAPSHOT_SSDATE (this should not happen"\
        "and, therefore, you should not be reading this message;"\
        "it probably means either that an external agent has disturbed"\
        "program's temporary runtime data files or that there is"\
        "a bug in the program)"
    }
    SNAPSHOT_SSDATE="$PHB_SNAPSHOT1"
    SNAPSHOT_D2DATE="$(convert_format_ss_to_d2 "$PHB_SNAPSHOT1")"
    touch -d "$SNAPSHOT_D2DATE" "$TMPDIR/snapshot.stamp"
    (($OPT_STAMP)) && touch -d "$SNAPSHOT_D2DATE" "$STAMPOPT_FILE"
    true # necessary because the last line has &&

}

# IMPLEMENTATION NOTES
#
# The question of scripting languages has been raised. Bash has in fact been
# used to implement not only this script but the whole program, so the question
# is perhaps moot, but one still might ask: why Bash?
#
# No full justification of the choice of will be attempted but a few
# notes follow.
#
# Had the author followed his personal taste, he'd probably have written the
# program in Perl; but then future maintainers would have had to learn Perl,
# wouldn't they? If you know and like Perl then you might think that a capital
# idea, but there are others who would disagree. Moreover, best use of Perl for
# tasks like the task this program undertakes involves the invocation of
# relatively obscure Perl modules. Such modules are great but then the future
# maintainer has to learn the modules, too. These are not minor issues.
# Consider how often C is still used in preference to C++ (though this writer
# happens to prefer C++) for little other reason than that C and its standard
# library are *small.* Insofar as a reader of a program written in a small
# language can reasonably be expected to understand the immediate operation
# of *any line of code* written in conventional style in the small language,
# small languages are readable in a way that large languages are not and
# cannot be.
#
# If one means to maintain just about anything of this program's kind, then one
# ought at least to know or learn Bash. Such an expectation is more reasonable
# than an expectation that one ought to know or learn Perl, C++ or the like.
# If you are, say, a Python programmer unfamiliar with Bash, then you probably
# wanted to learn Bash, anyway. (If for some reason you didn't, then
# maintaining this program is probably not for you.)
#
# Of course, the program could have been written in C, instead, which is as
# small and familiar as Bash is. A C program would have run quicker. However, C
# was not designed especially to move files around, whereas Bash was indeed
# designed especially to do just this. As far as quick goes, when a program is
# downloading gigabytes of files from remote servers as this program does, it
# is not clear that it matters very much whether a Bash script take an extra
# few seconds to process something or other on the local CPU.
#
# Bash brings the extra developmental advantage that, though Bash does support
# rudimentary arrays, Bash encourages the storing of lists of data instead in
# files in a temporary working directory, which can alternately be cleaned
# up (using LEAVE_TEMP_DATA_BEHIND=0 in the runtime configuration file) or left
# behind (using =1) when the program exits. Though the latter option does not
# wholly offset Bash's relative lack of overall structure, the option is
# nevertheless neat and convenient for debugging. Of course, other languages
# can serialize data for writeout, too; but in Bash the data is already
# serialized and written out.
#
# To emphasize the point: Bash's familiarity is especially important to this
# particular program because the program is expected to require substantial
# updates upon each major release of Debian. If the original author is
# unavailable to make the update, then the update must be feasible for a
# competent but as-yet unknown future maintainer (with as-yet unknown skills)
# to make.
#
# Attacking from a different angle, someone might alternately object, "Well,
# then, if it's smallness and universality you want, then why Bash? Why not
# the POSIX subset of Bash?"
#
# This is an objection of a different nature which applies to installation
# scripts but is probably specious here. The question here is not whether Bash
# is POSIX, nor whether Bash is theoretically universal, but whether Bash is
# familiar. The POSIX objection is thus rejected as irrelevantly grounded.
# Every Debian user has Bash installed on his or her system and, as far as this
# writer knows, most Debian users regularly use Bash to the extent to which
# they use any shell at all. For a Debian user to distinguish Bashisms from
# pure POSIXisms would be a vexing chore in itself! It is also an unnecessary
# chore. POSIX is more prolix than Bash and is insufficiently smaller than Bash
# to compensate for the prolixity. This program is no installation script but
# an already installed program, running on an already configured system. Bash
# is used.
#
# The one tool the program employs with which future maintainers might be
# unfamiliar is Sed. Sed is quite standard, of course, but many prefer the
# equally fine, and equally standard, Awk. The original author had to make a
# choice. Sed scripts are harder to read but Sed's manual is shorter than
# Awk's. On the ground that, if a future maintainer had to go to read a manual,
# the shorter manual were preferable to the longer, Sed has been employed.
# (Also, Awk is wedged between Sed and Perl in terms of its level of
# capability. Awk is too close to Sed for Sed users to bother to learn Awk, on
# the one side, and too close to Perl for Perl users to bother to learn Awk, on
# the other. The author happens to be a Sed/Perl user who, thus, is admittedly
# less familiar with Awk than with Sed. The main point however is the manual's
# length. The author would have used Awk if Awk's manual had been substantially
# the shorter. If you are an Awk user going to read Sed's manual to maintain
# this program, then start by reading the manual's half-page-long
# subsection, "Programming Commands," where Sed's 'b' and 't' commands are
# explained. Then read "Common Commands," which is about a page in length.
# GNU Sed's 'B' and 'T' commands are opposites of 'b' and 't'. Sed demarcates
# regexes either by /REGEX/ or by \%REGEX% -- the latter style being convenient
# when the REGEX happens to contain slashes. These should get you started at
# maintenance without using too much of your time.)

